// Structures to load assets from files.
PartyLoader partyLoader;
RoomLoader roomLoader;
// GUI stuff.
ArrayList<CharacterHUD> huds;
MainTextBox mainTextBox;
int scrollX, scrollY;
int deadZoneMinX, deadZoneMaxX, deadZoneMinY, deadZoneMaxY;  // Defines a box in screen coordinates where characters can move without auto-scrolling the screen.
// Tracks rooms that have been discovered.
ArrayList<RoomData> discoveredRooms;
// Create Image Objects
// TODO: Get real dead sprite image.
PImage deadSprite;
PImage doorN, doorS, doorE, doorW;

// Setup Player info.
int startingPartySize;
ArrayList<CharacterData> characters;
int moveSpeed, scrollSpeed;
boolean characterMoved;  // Used as a check when redrawing the screen.

// DEBUG.
boolean DEBUG = true;
boolean invertedKeyboardScrolling = true;  // Controls how scrolling the screen with the keyboard performs.

void setup() {
  size(1280, 720);
  // Initialize global vars.
  deadZoneMinX = 200;
  deadZoneMaxX = 900;
  deadZoneMinY = 150;
  deadZoneMaxY = 550;
  
  // Load game assets, prep the GUI.
  startingPartySize = 4;
  characterMoved = false;
  moveSpeed = 5;
  scrollSpeed = 10;
  mainTextBox = new MainTextBox(mainTextBoxLeft, mainTextBoxTop, mainTextBoxWidth, mainTextBoxHeight, mainFontSize, mainTextBoxStroke, mainTextBoxFill);
  loadData();
  discoveredRooms = new ArrayList<RoomData>();
  
  // Load static assets to image objects.
  deadSprite = loadImage("Sprites/Sprite Placeholder.png");
  doorN = loadImage("Door Placeholder.png");
  doorE = loadImage("Door Placeholder 2.png");
  doorW = loadImage("Door Placeholder 3.png");
  doorS = loadImage("Door Placeholder 4.png");
  
  // Kick off a new game.
  startGame();
}

void draw() {

  update();

  //DRAW
  // Reset the view and start drawing anew.
  background(64);

  // Render rooms.
  for(int i = 0; i < discoveredRooms.size(); i++) {
    discoveredRooms.get(i).drawFloor();
  }

  // Render players.
  for(int i = 0; i < characters.size(); i++) {
    characters.get(i).draw();
  }
  
  // Render walls etc.
  for(int i = 0; i < discoveredRooms.size(); i++) {
    discoveredRooms.get(i).drawWallAndDoors();
  }
  
  // Draw all the character HUDs.
  for(int i = 0; i < huds.size(); i++) {
    huds.get(i).draw();
  }
  // Draw the border around the character HUDs.
  noFill();
  stroke(hudBorderColor);
  rect(hudLeft - hudPadding, hudTop - hudPadding, hudWidth + 2 * hudPadding, (hudHeight * 4) + (5 * hudPadding), 5);
  
  // Draw the main text box if applicable.
  if(mainTextBox.displayingText) {
    mainTextBox.redraw();
  }
}

void keyPressed() {
  // Screen scrolling.
  int tempScrollX = 0;
  int tempScrollY = 0;
  if (key == 'w') {
    tempScrollY = -1 * scrollSpeed;
  }
  if (key == 's') {
    tempScrollY = scrollSpeed;
  }
  if (key == 'a') {
    tempScrollX = -1 * scrollSpeed;
  }
  if (key == 'd') {
    tempScrollX = scrollSpeed;
  }
  if (invertedKeyboardScrolling) {
    scrollX -= tempScrollX;
    scrollY -= tempScrollY;
  } else {
    scrollX += tempScrollX;
    scrollY += tempScrollY;
  }
  // Character selection & moving.
  if(key >= '1' && key <= '4') {
    selectActiveCharacter(key - '1');
  }
  if (key == CODED) {
    if (keyCode == UP) {
      moveActiveCharacter(0, -1 * moveSpeed);
    } else if (keyCode == DOWN) {
      moveActiveCharacter(0, moveSpeed);
    }
    if (keyCode == LEFT) {
      moveActiveCharacter(-1 * moveSpeed, 0);
    } else if (keyCode == RIGHT) {
      moveActiveCharacter(moveSpeed, 0);
    }
  }
  if (key == ' ') {
    endTurn();
  }
  
  if(DEBUG) {
    // HUD debug-related stuff.
    if(getActiveCharacterIndex() >= 0) {
      if(key == '=' || key == '+') {
        characters.get(getActiveCharacterIndex()).changeHealth(1);
      } else if(key == '-') {
        characters.get(getActiveCharacterIndex()).changeHealth(-1);
      } else if(key == '[') {
        characters.get(getActiveCharacterIndex()).moveLeft -= 1;
      } else if(key == ']') {
        characters.get(getActiveCharacterIndex()).moveLeft += 1;
      }
    }
    // Text box debug-related stuff.
    if(key == ENTER && mainTextBox.displayingText == true) {
      mainTextBox.closeText();
    } else if(key == ENTER && mainTextBox.displayingText == false) {
      mainTextBox.displayText("And then there were none...", scaryColor);
    }
  }
}

void mousePressed() {
  // Select an active character if their portrait was clicked on.
  for(int i = 0; i < huds.size(); i++) {
    if(huds.get(i).isMouseOver()) {
      selectActiveCharacter(i);
    }
  }
}

void update() {
  // If the (presumably active) character moved, check if the screen needs to adjust to show that.
  if(characterMoved) {
    if(getActiveCharacterIndex() >= 0) {
      checkScreenScroll(characters.get(getActiveCharacterIndex()));
    }
    characterMoved = false;
  }
}

/* Checks if the scrren needs to scroll to put/keep the active character visible. */
void checkScreenScroll(CharacterData characterRef) {
  int screenX = characterRef.x + scrollX;
  int screenY = characterRef.y + scrollY;
  if(screenX < deadZoneMinX) {
    scrollX = -1 * characterRef.x + deadZoneMinX;
  } else if(screenX + characterWidth > deadZoneMaxX) {
    scrollX = -1 * (characterRef.x + characterWidth) + deadZoneMaxX;
  }
  if(screenY < deadZoneMinY) {
    scrollY = -1 * characterRef.y + deadZoneMinY;
  } else if(screenY + characterHeight > deadZoneMaxY) {
    scrollY = -1 * (characterRef.y + characterHeight) + deadZoneMaxY;
  }
}

/* Load up game data from the text files. */
void loadData() {
  partyLoader = new PartyLoader("party.csv");
  for(int i = 0; i < partyLoader.size(); i++) {
    println(partyLoader.getPartyMemberSummary(i));
  }
  roomLoader = new RoomLoader("rooms.csv");
  for(int i = 0; i < roomLoader.size(); i++) {
    println(roomLoader.getRoomSummary(i));
  }
}

/* Generate character information and their HUDs. */
void createCharacters() {
  HashMap<Integer,Integer> indexes = new HashMap<Integer,Integer>();
  // Choose four random (unique) characters from the list that was loaded.
  while(indexes.keySet().size() < startingPartySize) {
    int newIndex = int(random(partyLoader.size()));
    indexes.put(newIndex, newIndex);
  }
  characters = new ArrayList<CharacterData>();
  huds = new ArrayList<CharacterHUD>();
  int i = 0;
  for(int index : indexes.keySet()) {
    characters.add(new CharacterData(partyLoader.getPartyMember(index)));
    huds.add(new CharacterHUD(hudLeft, hudTop + ((hudHeight + hudPadding) * i), partyLoader.getPartyMember(index), characters.get(characters.size() - 1)));
    i++;
  }
}

/* Sets the passed character index to active and unselects the others. */
void selectActiveCharacter(int index) {
  for(int i = 0; i < characters.size(); i++) {
    if(i != index) {
      characters.get(i).isActive = false;
    } else {
      characters.get(i).isActive = true;
    }
  }
}

/* Kicks off a new game. */
void startGame() {
  // Create a new, random party of characters.
  createCharacters();
  // Always start with the bare room that's first in the rooms list.
  discoveredRooms = new ArrayList<RoomData>();
  discoveredRooms.add(new RoomData(roomLoader.getRoomInfo(0), 0, 0));
  // Stick the characters randomly in the first room.
  for(int i = 0; i < characters.size(); i++) {
    characters.get(i).putInRoom(discoveredRooms.get(0));
  }
  // Start (0,0) in the center(-ish) of the screen.
  scrollX = (width / 2) - sizeUnitScale;
  scrollY = (height / 2) - sizeUnitScale;
}

/* Run logic to kick off a new player "turn". */
void endTurn() {
  // Reset everyone's movement points.
  for(int i = 0; i < characters.size(); i++) {
    characters.get(i).changeMovement(characters.get(i).maxMove);
  }
  // TODO: This would be where "end of turn" house events would happen.
}
