int mainTextBoxLeft = 10;
int mainTextBoxTop = 640;
int mainTextBoxWidth = 1100;
int mainTextBoxHeight = 40;
int textBoxPadding = 4;
color mainTextBoxStroke = color(0, 255, 255);
color mainTextBoxFill = color(128, 255, 255, 128);

int mainFontSize = 24;
color normalColor = color(0, 0, 32);
color scaryColor = color(64, 0, 16);

PFont mainTextFont = createFont("Georgia", mainFontSize, true);

/* A UI element to display text in a box. */
class TextBox {
  int x;
  int y;
  int w;
  int h;
  int fontSize;
  color strokeCol;  // The color of the border of the text box.
  color fillCol;  // The color of the "background" of the text box.
  
  TextBox(int _x, int _y, int _w, int _h, int _fontSize, color _strokeCol, color _fillCol) {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    fontSize = _fontSize;
    strokeCol = _strokeCol;
    fillCol = _fillCol;
  }
  
  /* Displays the passed text in a text box. */
  void displayText(String dispText, color col) {
    // Draw the text box.
    tint(255);
    stroke(strokeCol);
    fill(fillCol);
    rect(x, y, w, h);
    // Render the text on top of the text box.
    textFont(mainTextFont, fontSize);
    fill(col);
    text(dispText, x + textBoxPadding, y + fontSize + textBoxPadding);
  }
}

/* Subclass of TextBox intended to be the "main" info text box of the GUI. */
class MainTextBox extends TextBox {
  PImage screenSave;  // Used to preserve what the window looked like before the box popped up.
  boolean displayingText;
  String currentText;
  color currentCol;
  
  MainTextBox(int _x, int _y, int _w, int _h, int _fontSize, color _strokeCol, color _fillCol) {
    super(_x, _y, _w, _h, _fontSize, _strokeCol, _fillCol);
    screenSave = createImage(width, height, RGB);
    displayingText = false;
    currentText = "";
    currentCol = color(0, 0, 0);
  }
  
  /* Pops up the passed text in a text box. */
  void displayText(String dispText, color col) {
    // Only save the background if this is the first frame this text box is visible.
    if(displayingText == false) {
      savePixels(screenSave);
    }
    currentText = dispText;
    currentCol = col;
    displayingText = true;
    redraw();  // Easier to just call this rather than repeating the code here.
  }
  
  /* Re-renders the text box without changing anything. */
  void redraw() {
    // Draw the text box.
    tint(255);
    stroke(strokeCol);
    fill(fillCol);
    rect(x, y, w, h, 7);
    // Render the text on top of the text box.
    textFont(mainTextFont, fontSize);
    fill(currentCol);
    text(currentText, x + textBoxPadding, y + fontSize + textBoxPadding);
  }
  
  /* Closes the text box and restores what was underneath. */
  void closeText() {
    restorePixels(screenSave);
    displayingText = false;
  }
  
}

/* Saves off the current screen so it can be restored later. */
void savePixels(PImage buffer) {
  buffer.loadPixels();
  loadPixels();
  int loc;
  for(int x = 0; x < width; x++) {
    for(int y = 0; y < height; y++) {
      loc = x + y * width;
      buffer.pixels[loc] = pixels[loc];
    }
  }
  updatePixels();
  buffer.updatePixels();
}

/* Restores the screen to what was saved in a PImage. */
void restorePixels(PImage buffer) {
  buffer.loadPixels();
  loadPixels();
  int loc;
  for(int x = 0; x < width; x++) {
    for(int y = 0; y < height; y++) {
      loc = x + y * width;
      pixels[loc] = buffer.pixels[loc];
    }
  }
  updatePixels();
  buffer.updatePixels();
}
