import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class Main extends PApplet {

// Structures to load assets from files.
PartyLoader partyLoader;
RoomLoader roomLoader;
// GUI stuff.
ArrayList<CharacterHUD> huds;
MainTextBox mainTextBox;
int scrollX, scrollY;
int deadZoneMinX, deadZoneMaxX, deadZoneMinY, deadZoneMaxY;  // Defines a box in screen coordinates where characters can move without auto-scrolling the screen.
// Tracks rooms that have been discovered.
ArrayList<RoomData> discoveredRooms;
// Create Image Objects
// TODO: Get real dead sprite image.
PImage deadSprite;
PImage doorN, doorS, doorE, doorW;

// Setup Player info.
int startingPartySize;
ArrayList<CharacterData> characters;
int moveSpeed, scrollSpeed;
boolean characterMoved;  // Used as a check when redrawing the screen.

// DEBUG.
boolean DEBUG = true;
boolean invertedKeyboardScrolling = true;  // Controls how scrolling the screen with the keyboard performs.

public void setup() {
  size(1280, 720);
  // Initialize global vars.
  deadZoneMinX = 200;
  deadZoneMaxX = 900;
  deadZoneMinY = 150;
  deadZoneMaxY = 550;
  
  // Load game assets, prep the GUI.
  startingPartySize = 4;
  characterMoved = false;
  moveSpeed = 5;
  scrollSpeed = 10;
  mainTextBox = new MainTextBox(mainTextBoxLeft, mainTextBoxTop, mainTextBoxWidth, mainTextBoxHeight, mainFontSize, mainTextBoxStroke, mainTextBoxFill);
  loadData();
  discoveredRooms = new ArrayList<RoomData>();
  
  // Load static assets to image objects.
  deadSprite = loadImage("Sprites/Sprite Placeholder.png");
  doorN = loadImage("Door Placeholder.png");
  doorE = loadImage("Door Placeholder 2.png");
  doorW = loadImage("Door Placeholder 3.png");
  doorS = loadImage("Door Placeholder 4.png");
  
  // Kick off a new game.
  startGame();
}

public void draw() {

  update();

  //DRAW
  // Reset the view and start drawing anew.
  background(64);

  // Render rooms.
  for(int i = 0; i < discoveredRooms.size(); i++) {
    discoveredRooms.get(i).drawFloor();
  }

  // Render players.
  for(int i = 0; i < characters.size(); i++) {
    characters.get(i).draw();
  }
  
  // Render walls etc.
  for(int i = 0; i < discoveredRooms.size(); i++) {
    discoveredRooms.get(i).drawWallAndDoors();
  }
  
  // Draw all the character HUDs.
  for(int i = 0; i < huds.size(); i++) {
    huds.get(i).draw();
  }
  // Draw the border around the character HUDs.
  noFill();
  stroke(hudBorderColor);
  rect(hudLeft - hudPadding, hudTop - hudPadding, hudWidth + 2 * hudPadding, (hudHeight * 4) + (5 * hudPadding), 5);
  
  // Draw the main text box if applicable.
  if(mainTextBox.displayingText) {
    mainTextBox.redraw();
  }
}

public void keyPressed() {
  // Screen scrolling.
  int tempScrollX = 0;
  int tempScrollY = 0;
  if (key == 'w') {
    tempScrollY = -1 * scrollSpeed;
  }
  if (key == 's') {
    tempScrollY = scrollSpeed;
  }
  if (key == 'a') {
    tempScrollX = -1 * scrollSpeed;
  }
  if (key == 'd') {
    tempScrollX = scrollSpeed;
  }
  if (invertedKeyboardScrolling) {
    scrollX -= tempScrollX;
    scrollY -= tempScrollY;
  } else {
    scrollX += tempScrollX;
    scrollY += tempScrollY;
  }
  // Character selection & moving.
  if(key >= '1' && key <= '4') {
    selectActiveCharacter(key - '1');
  }
  if (key == CODED) {
    if (keyCode == UP) {
      moveActiveCharacter(0, -1 * moveSpeed);
    } else if (keyCode == DOWN) {
      moveActiveCharacter(0, moveSpeed);
    }
    if (keyCode == LEFT) {
      moveActiveCharacter(-1 * moveSpeed, 0);
    } else if (keyCode == RIGHT) {
      moveActiveCharacter(moveSpeed, 0);
    }
  }
  if (key == ' ') {
    endTurn();
  }
  
  if(DEBUG) {
    // HUD debug-related stuff.
    if(getActiveCharacterIndex() >= 0) {
      if(key == '=' || key == '+') {
        characters.get(getActiveCharacterIndex()).changeHealth(1);
      } else if(key == '-') {
        characters.get(getActiveCharacterIndex()).changeHealth(-1);
      } else if(key == '[') {
        characters.get(getActiveCharacterIndex()).moveLeft -= 1;
      } else if(key == ']') {
        characters.get(getActiveCharacterIndex()).moveLeft += 1;
      }
    }
    // Text box debug-related stuff.
    if(key == ENTER && mainTextBox.displayingText == true) {
      mainTextBox.closeText();
    } else if(key == ENTER && mainTextBox.displayingText == false) {
      mainTextBox.displayText("And then there were none...", scaryColor);
    }
  }
}

public void mousePressed() {
  // Select an active character if their portrait was clicked on.
  for(int i = 0; i < huds.size(); i++) {
    if(huds.get(i).isMouseOver()) {
      selectActiveCharacter(i);
    }
  }
}

public void update() {
  // If the (presumably active) character moved, check if the screen needs to adjust to show that.
  if(characterMoved) {
    if(getActiveCharacterIndex() >= 0) {
      checkScreenScroll(characters.get(getActiveCharacterIndex()));
    }
    characterMoved = false;
  }
}

/* Checks if the scrren needs to scroll to put/keep the active character visible. */
public void checkScreenScroll(CharacterData characterRef) {
  int screenX = characterRef.x + scrollX;
  int screenY = characterRef.y + scrollY;
  if(screenX < deadZoneMinX) {
    scrollX = -1 * characterRef.x + deadZoneMinX;
  } else if(screenX + characterWidth > deadZoneMaxX) {
    scrollX = -1 * (characterRef.x + characterWidth) + deadZoneMaxX;
  }
  if(screenY < deadZoneMinY) {
    scrollY = -1 * characterRef.y + deadZoneMinY;
  } else if(screenY + characterHeight > deadZoneMaxY) {
    scrollY = -1 * (characterRef.y + characterHeight) + deadZoneMaxY;
  }
}

/* Load up game data from the text files. */
public void loadData() {
  partyLoader = new PartyLoader("party.csv");
  for(int i = 0; i < partyLoader.size(); i++) {
    println(partyLoader.getPartyMemberSummary(i));
  }
  roomLoader = new RoomLoader("rooms.csv");
  for(int i = 0; i < roomLoader.size(); i++) {
    println(roomLoader.getRoomSummary(i));
  }
}

/* Generate character information and their HUDs. */
public void createCharacters() {
  HashMap<Integer,Integer> indexes = new HashMap<Integer,Integer>();
  // Choose four random (unique) characters from the list that was loaded.
  while(indexes.keySet().size() < startingPartySize) {
    int newIndex = PApplet.parseInt(random(partyLoader.size()));
    indexes.put(newIndex, newIndex);
  }
  characters = new ArrayList<CharacterData>();
  huds = new ArrayList<CharacterHUD>();
  int i = 0;
  for(int index : indexes.keySet()) {
    characters.add(new CharacterData(partyLoader.getPartyMember(index)));
    huds.add(new CharacterHUD(hudLeft, hudTop + ((hudHeight + hudPadding) * i), partyLoader.getPartyMember(index), characters.get(characters.size() - 1)));
    i++;
  }
}

/* Sets the passed character index to active and unselects the others. */
public void selectActiveCharacter(int index) {
  for(int i = 0; i < characters.size(); i++) {
    if(i != index) {
      characters.get(i).isActive = false;
    } else {
      characters.get(i).isActive = true;
    }
  }
}

/* Kicks off a new game. */
public void startGame() {
  // Create a new, random party of characters.
  createCharacters();
  // Always start with the bare room that's first in the rooms list.
  discoveredRooms = new ArrayList<RoomData>();
  discoveredRooms.add(new RoomData(roomLoader.getRoomInfo(0), 0, 0));
  // Stick the characters randomly in the first room.
  for(int i = 0; i < characters.size(); i++) {
    characters.get(i).putInRoom(discoveredRooms.get(0));
  }
  // Start (0,0) in the center(-ish) of the screen.
  scrollX = (width / 2) - sizeUnitScale;
  scrollY = (height / 2) - sizeUnitScale;
}

/* Run logic to kick off a new player "turn". */
public void endTurn() {
  // Reset everyone's movement points.
  for(int i = 0; i < characters.size(); i++) {
    characters.get(i).changeMovement(characters.get(i).maxMove);
  }
  // TODO: This would be where "end of turn" house events would happen.
}
// The "hitbox" and size of each character sprite.
int characterWidth = 60;
int characterHeight = 60;

/* Tracks data for a character in the game. */
class CharacterData {
  PImage sprite;
  int health;
  int maxHealth;
  int moveLeft;
  int maxMove;
  boolean isDead;
  int x;
  int y;
  boolean isActive;  // True if this is the character the player is controlling right now.
  RoomData curRoom;  // For reference, the RoomData structure representing where the character currently sits.
  
  CharacterData(PartyMemberInfo _info) {
    sprite = loadImage(_info.spritePicture);
    maxHealth = _info.maxHealth;
    health = maxHealth;
    maxMove = _info.maxMovement;
    moveLeft = maxMove;
    isDead = false;
    x = 0;
    y = 0;
    isActive = false;
    curRoom = null;
  }
  
  /* Draws the appropriate character sprite. Assumes that translation for scrolling has _not_ been performed. */
  public void draw() {
    if(isDead) {
      tint(128);
      image(deadSprite, x + scrollX, y + scrollY);
    } else {
      tint(255);
      image(sprite, x + scrollX, y + scrollY);
    }
  }
  
  /* Puts the character in a random position in the given room. */
  public void putInRoom(RoomData room) {
    int minX = room.unitsX * sizeUnitScale + characterWidth;
    int maxX = (room.unitsX + room.roomRef.unitWidth) * sizeUnitScale - characterWidth;
    int minY = room.unitsY * sizeUnitScale + characterHeight;
    int maxY = (room.unitsY + room.roomRef.unitHeight) * sizeUnitScale - characterHeight;
    x = PApplet.parseInt(random(minX, maxX));
    y = PApplet.parseInt(random(minY, maxY));
    curRoom = room;
  }
  
  /* Puts the character in the given room next to the given door. */
  public void putInRoomByDoor(RoomData room, DoorInfo door) {
    curRoom = room;
    // TODO: Replace this with real stuff.
    x = room.unitsX * sizeUnitScale + (room.roomRef.unitWidth * sizeUnitScale / 2);
    y = room.unitsY * sizeUnitScale + (room.roomRef.unitHeight * sizeUnitScale / 2);
  }
  
  /* Increase or decrease the character's health by the given value, react accordingly. */
  public void changeHealth(int healthDelta) {
    health += healthDelta;
    if(health < 1) {
      isDead = true;
    } else {
      isDead = false;
    }
  }
  
  /* Increase or decrease the character's movement points by the given value, react accordingly. 
     Returns true if the value of moveLeft actually changed.*/
  public boolean changeMovement(int moveDelta) {
    int oldMove = moveLeft;
    moveLeft += moveDelta;
    if(moveLeft < 0) {
      moveLeft = 0;
    } else if(moveLeft > maxMove) {
      moveLeft = maxMove;
    }
    return oldMove != moveLeft;
  }
  
  /* Try to shift the character's position by the given values.
     Returns true if the character actually moved. */
  public boolean shift(int _xShift, int _yShift) {
    boolean moved = false;
    boolean madeNewRoom = false;  // Used to track potential event kick-offs.
    int collideDoor = -1;
    if(moveLeft > 0) {
      x += _xShift;
      y += _yShift;
      // Check for a door collision.
      collideDoor = checkDoorCollision();
      if(collideDoor >= 0) {
        // Try to find a valid existing room on the other side of the door or make a new one.
        int newRoomIndex = -1;
        int newDoorIndex = -1;
        newRoomIndex = getNeighborRoom(curRoom, curRoom.roomRef.doors.get(collideDoor));
        if(newRoomIndex >= 0) {
          // Confirm that the room on the other side of the door has a connecting door.
          newDoorIndex = getNextRoomDoor(curRoom, discoveredRooms.get(newRoomIndex), curRoom.roomRef.doors.get(collideDoor));
          if(newDoorIndex >= 0) {
            curRoom = discoveredRooms.get(newRoomIndex);
            putInRoomByDoor(curRoom, curRoom.roomRef.doors.get(newDoorIndex));
            moved = true;
          }
        } else {
          // If there's not a neighboring room yet, make one.
            newRoomIndex = generateNewRoom(curRoom, curRoom.roomRef.doors.get(collideDoor));
            if(newRoomIndex >= 0) {
              // Confirm that the room on the other side of the door has a connecting door.
              newDoorIndex = getNextRoomDoor(curRoom, discoveredRooms.get(newRoomIndex), curRoom.roomRef.doors.get(collideDoor));
              if(newDoorIndex >= 0) {
                curRoom = discoveredRooms.get(newRoomIndex);
                putInRoomByDoor(curRoom, curRoom.roomRef.doors.get(newDoorIndex));
                moved = true;
              } else {
                // NOTE: Weird state, made a new (supposedly valid) room, but no connecting door found. If it comes up often, handle appropriately.
              }
              madeNewRoom = true;
            } else {
              // NOTE: Weird state, making a new room failed. If this comes up often, add code to handle appropriately.
            }
        }
        if(madeNewRoom) {
          // TODO: This is where "discovered a new room" events would potentially kick off. Call a function to see if one happens, and then execute one if so.
        }
      } else {  
        // Check for a wall collision if no door collision.
        if(checkWallCollision()) {
          x -= _xShift;
          y -= _yShift;
        } else {
          moved = true;
        }
      }
      // If, after all that, the character actually moved, decrement their move points.
      if(moved) {
        changeMovement(-1);
      }
    }
    return moved;
  }
  
  /* Checks for a collision between the character and any wall of the room it's in. */ 
  public boolean checkWallCollision() {
    boolean collision = false;
    if(curRoom != null) {
      int roomMinX = curRoom.unitsX * sizeUnitScale;
      int roomMaxX = (curRoom.unitsX + curRoom.roomRef.unitWidth) * sizeUnitScale;
      int roomMinY = curRoom.unitsY * sizeUnitScale;
      int roomMaxY = (curRoom.unitsY + curRoom.roomRef.unitHeight) * sizeUnitScale;
      if(x < roomMinX || x + characterWidth > roomMaxX || y < roomMinY || y + characterHeight > roomMaxY) {
        collision = true;
      }
    }
    return collision;
  }
  
  /* Checks if the character has "collided" with any of the doors in its current room.
     Returns -1 if no, or the index into the room's doors array of the door that collided. */
  public int checkDoorCollision() {
    int collidedDoor = -1;
    if(curRoom != null) {
      collidedDoor = curRoom.checkCharacterInDoor(this);
    }
    return collidedDoor;
  }
}

/* Adjusts the active character's position by the given values. */
public void moveActiveCharacter(int _xShift, int _yShift) {
  for(int i = 0; i < characters.size(); i++) {
    if(characters.get(i).isActive && characters.get(i).moveLeft > 0) {
      characterMoved = characters.get(i).shift(_xShift, _yShift);
    }
  }
}

/* Finds and returns the index of the first active character, or -1 if there are none. */
public int getActiveCharacterIndex() {
  for(int i = 0; i < characters.size(); i++) {
    if(characters.get(i).isActive) {
      return i;
    }
  }
  return -1;
}
int hudWidth = 100;
int nameBoxFontSize = 16;
int nameBoxInternalPadding = 4;  // Padding between the text in each name box and its border.
int nameBoxWidth = hudWidth;
int nameBoxHeight = nameBoxFontSize + (nameBoxInternalPadding * 2);
int nameBoxFill = color(128, 255, 255);
int nameBoxBorder = nameBoxFill;
int nameBoxTextColor = color(0, 0, 0);
int portraitWidth = hudWidth;
int portraitHeight = 100;
int healthBarWidth = hudWidth;
int healthBarHeight = 10;
int movementBarWidth = hudWidth;
int movementBarHeight = 10;
int maxMovementPoints = 100;  // The most movement points any character can have.
int hudElementPadding = 3;  // Vertical space between HUD elements.
int hudHeight = nameBoxHeight + portraitHeight + healthBarHeight + movementBarHeight + (3 * hudElementPadding);

int mouseOverColor = color(0, 255, 0, 64);
int selectedColor = color(128, 255, 128, 64);

int hudLeft = 1150;  // The x of the character HUDs.
int hudTop = 20;  // The initial top of the first character HUD.
int hudPadding = 10;  // Space between each character HUD.
int hudBorderColor = color(0, 255, 255);

/* Represents the Heads-Up Display for a character. */
class CharacterHUD {
  int x;
  int y;
  int w;  // Used for "is the mouse over me?" checks.
  int h;  // Used for "is the mouse over me?" checks.
  TextBox nameBox;
  CharacterPortrait portrait;
  HealthBar healthBar;
  MovementBar moveBar;
  String characterName;
  CharacterData characterRef;  // The "real" data associated with this character.
  
  CharacterHUD(int _x, int _y, PartyMemberInfo _info, CharacterData _char) {
    x = _x;
    y = _y;
    w = hudWidth;
    h = hudHeight;
    characterName = _info.name;
    characterRef = _char;
    
    nameBox = new TextBox(x, y, nameBoxWidth, nameBoxHeight, nameBoxFontSize, nameBoxBorder, nameBoxFill);
    portrait = new CharacterPortrait(x, y + nameBoxHeight + hudElementPadding, _info);
    healthBar = new HealthBar(x, y + nameBoxHeight + portraitHeight + (hudElementPadding * 2), _info.maxHealth, _info.maxHealth);
    moveBar = new MovementBar(x, y + nameBoxHeight + portraitHeight + healthBarHeight + (hudElementPadding * 3), _info.maxMovement, _info.maxMovement);
  }
  
  /* Sets the display value for the health of a character. Also adjusts portrait picture. */
  public void setHealth(int newHealth) {
    healthBar.current = newHealth;
    if(newHealth < healthBar.max) {
      portrait.isHurt = true;
    } else {
      portrait.isHurt = false;
    }
    if(newHealth < 1) {
      portrait.isDead = true;
    } else {
      portrait.isDead = false;
    }
  }
  
  /* Sets the display value for the movement left of a character. */
  public void setMovement(int newMovement) {
    moveBar.current = newMovement;
  }
  
  /* Draws each HUD element. */
  public void draw() {
    setHealth(characterRef.health);
    setMovement(characterRef.moveLeft);
    nameBox.displayText(characterName, nameBoxTextColor);
    portrait.draw();
    healthBar.draw();
    moveBar.draw();
    // Draw extra stuff if the mouse is over the HUD or it's selected.
    if(isMouseOver() == true && characterRef.isActive == false) {
      noStroke();
      fill(mouseOverColor);
      rect(x - (hudPadding / 2), y - (hudPadding / 2), w + hudPadding, h + hudPadding);
    } else if(characterRef.isActive) {
      noStroke();
      fill(selectedColor);
      rect(x - (hudPadding / 2), y - (hudPadding / 2), w + hudPadding, h + hudPadding);
    }
  }
  
  /* Checks if the mouse is over any part of this HUD element. */
  public boolean isMouseOver() {
    return (mouseX >= x && mouseX <= x + w && mouseY >= y && mouseY <= y + h);
  }
}

/* Represents a character's portrait as part of the "HUD" for that character. */
class CharacterPortrait {
  int x;
  int y;
  int w;
  int h;
  PImage healthyImage;
  PImage hurtImage;
  PImage deadImage;
  boolean isHurt;
  boolean isDead;
  
  CharacterPortrait(int _x, int _y, PartyMemberInfo _info) {
    x = _x;
    y = _y;
    w = portraitWidth;
    h = portraitHeight;
    isHurt = false;
    isDead = false;
    healthyImage = loadImage(_info.healthyPicture);
    hurtImage = loadImage(_info.hurtPicture);
    deadImage = loadImage(_info.deadPicture);
  }
  
  public void draw() {
    tint(255);
    if(isDead == true) {
      image(deadImage, x, y, w, h);
    } else if(isHurt == true) {
      image(hurtImage, x, y, w, h);
    } else {
      image(healthyImage, x, y, w, h);
    }
  }
}

/* Represents a health bar to display under a character's portrait. */
class HealthBar {
  int x;
  int y;
  int w;
  int h;
  int max;  // Maximum health points.
  int current;  // Current health points.
  // Size of each "tick" on the health bar.
  int tickWidth;
  int tickHeight;
  int tickSpace;
  int borderCol;
  int healthCol;
  int hurtCol;
  
  HealthBar(int _x, int _y, int _max, int _current) {
    x = _x;
    y = _y;
    w = healthBarWidth;
    h = healthBarHeight;
    max = _max;
    current = _current;
    
    tickWidth = 30;
    tickHeight = 10;
    tickSpace = 5;
    borderCol = color(128, 0, 0);
    healthCol = color(255, 0, 0);
    hurtCol = color(0, 0, 0);
  }
  
  /* Draw a series of health "ticks". */
  public void draw() {
    stroke(borderCol);
    for(int i = 0; i < max; i++) {
      if(i + 1 <= current) {
        fill(healthCol);
      } else {
        fill(hurtCol);
      }
      rect(x + (i * (tickWidth + tickSpace)), y, tickWidth, tickHeight);
    }
  }
}

/* Represents a "movement left" bar to display under a character's portrait. */
class MovementBar {
  int x;
  int y;
  int w;
  int h;
  int max;  // Maximum movement points.
  int current;  // Current movement points.
  int borderCol;
  int barCol;  // The color to put in the "you have this much left" part of the bar.
  int usedBarCol;  // The color to put in the "you've used this much" part of the bar.
  int barPadding;  // How much space to put between between the border and the bar.

  MovementBar(int _x, int _y, int _max, int _current) {
    x = _x;
    y = _y;
    w = movementBarWidth;
    h = movementBarHeight;
    max = _max;
    current = _current;
    
    borderCol = color(255, 255, 255);
    barCol = color(255, 255, 0);
    usedBarCol = color(0, 0, 0);
  }
  
  /* Draws a horizontal colored bar representing movement left. */
  public void draw() {
    stroke(borderCol);
    fill(usedBarCol);
    int baseBarWidth = PApplet.parseInt(PApplet.parseFloat(max) / PApplet.parseFloat(maxMovementPoints) * w);
    int barWidth = PApplet.parseInt((PApplet.parseFloat(current) / PApplet.parseFloat(max)) * baseBarWidth);
    if(barWidth < 0) { barWidth = 0; }
    if(barWidth > w) { barWidth = w; }
    rect(x, y, baseBarWidth, h);
    fill(barCol);
    rect(x + barPadding, y + barPadding, barWidth - (barPadding * 2), h - (barPadding * 2));
  }
}


/* Loads data from a text file describing possible party members. */
class PartyLoader {
  ArrayList<PartyMemberInfo> partyInfo;
  
  PartyLoader(String fileName) {
    partyInfo = new ArrayList<PartyMemberInfo>();
    
    Table partyData = loadTable(fileName, "header");
    for(TableRow row : partyData.rows()) {
      String name = row.getString(0);
      String genderStr = row.getString(1);
      String healthyPicture = row.getString(2);
      String hurtPicture = row.getString(3);
      String deadPicture = row.getString(4);
      String spritePicture = row.getString(5);
      int maxHealth = row.getInt(6);
      int maxMovement = row.getInt(7);
      partyInfo.add(new PartyMemberInfo(name, genderStr, healthyPicture, hurtPicture, deadPicture, spritePicture, maxHealth, maxMovement));
    }
  }
  
  /* Creates a short text description of the party member. */
  public String getPartyMemberSummary(int index) {
    return partyInfo.get(index).getPartyMemberSummary();
  }
  
  public PartyMemberInfo getPartyMember(int index) {
    return partyInfo.get(index);
  }
  
  public int size() {
    return partyInfo.size();
  }
}

/* Container class for information loaded about a party member. */
class PartyMemberInfo {
  String name;
  int gender; // 0 = Male ("M"), 1 = Female ("F").
  String healthyPicture;
  String hurtPicture;
  String deadPicture;
  String spritePicture;
  int maxHealth;
  int maxMovement;
  
  PartyMemberInfo(String _name, String _gender, String _healthyPicture, String _hurtPicture, String _deadPicture, String _spritePicture, int _maxHealth, int _maxMovement) {
    name = _name;
    if(_gender.toLowerCase().equals("m")) {
      gender = 0;
    } else {
      gender = 1;
    }
    healthyPicture = _healthyPicture;
    hurtPicture = _hurtPicture;
    deadPicture = _deadPicture;
    spritePicture = _spritePicture;
    maxHealth = _maxHealth;
    maxMovement = _maxMovement;
  }
  
  /* Returns a human-readable gender for the party member. */
  public String getGenderName() {
    if(gender == 0) {
      return "Male";
    } else {
      return "Female";
    }
  }
  
  /* Creates a short text description of the party member. */
  public String getPartyMemberSummary() {
    String summary;
    summary = name;
    summary = summary + " (" + getGenderName() + ")";
    summary = summary + ", Max Health " + maxHealth;
    summary = summary + ", Speed " + maxMovement;
    return summary;
  }
}
// How many pixels to shift a door along its respective wall segment.
int doorOffset = 150;
int[] doorWidths = {40, 40, 20, 20};
int[] doorHeights = {70, 5, 40, 40};
int doorHitboxPadding = 3;  // How many pixels to extend the "hit box" of room doors.
int wallHeight = 80;

/* Class to hold information about "discovered" rooms. */
class RoomData {
  PImage floorImage;
  PImage wallImage;
  // The top-left position of the room in grid units.
  int unitsX;
  int unitsY;
  ArrayList<DoorInfo> doors;
  RoomInfo roomRef;
  
  RoomData(RoomInfo _room, int _unitsX, int _unitsY) {
    floorImage = loadImage(_room.floorPicture);
    wallImage = loadImage(_room.wallPicture);
    unitsX = _unitsX;
    unitsY = _unitsY;
    doors = _room.doors;
    roomRef = _room;
  }
  
  /* Assumes translation for scrolling has _not_ been done. */
  public void drawFloor() {
    // Draw the floor.
    if(isACharacterInRoom(this)) {
      tint(255);
    } else {
      tint(128);
    }
    image(floorImage, unitsX * sizeUnitScale + scrollX, unitsY * sizeUnitScale + scrollY);
  }
  
  /* Draw the walls and doors separately so they can end up "on top" of all other scenery.
     Assumes translation for scrolling has _not_ been done. */
  public void drawWallAndDoors() {
    // Draw the wall and door(s) for the room. Separated so it can be done after all floors are drawn.
    if(isACharacterInRoom(this)) {
      tint(255);
    } else {
      tint(128);
    }
    image(wallImage, unitsX * sizeUnitScale + scrollX, unitsY * sizeUnitScale - wallHeight + scrollY);
    // Draw any doors.
    for(int i = 0; i < doors.size(); i++) {
      int offset = doors.get(i).offset;
      switch(doors.get(i).side) {
        case 0: // North.
          // The "wallHeight - doorHeights..." compensates for the door not necessarily being as tall as the wall.
          image(doorN, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
        case 1: // South.
          image(doorS, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
        case 2: // East.
          image(doorE, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
        case 3: // West.
          image(doorW, getDoorX(doors.get(i)) + scrollX, getDoorY(doors.get(i)) + scrollY);
          break;
      }
    }
  }
  
  /* Checks if the given character is "in" any of this room's doors.
     Returns the index of the door in the doors array, or -1 if they're not colliding with any doors. */
  public int checkCharacterInDoor(CharacterData character) {
    int doorIndex = -1;
    int doorMinX, doorMaxX, doorMinY, doorMaxY; 
    int charMinX = character.x;
    int charMaxX = character.x + characterWidth;
    int charMinY = character.y;
    int charMaxY = character.y + characterWidth;
    for(int i = 0; i < doors.size(); i++) {
      DoorInfo curDoor = doors.get(i);
      int offset = curDoor.offset;
      doorMinX = getDoorX(curDoor) - doorHitboxPadding;
      doorMaxX = doorMinX + doorWidths[curDoor.side] + doorHitboxPadding;
      doorMinY = getDoorY(curDoor) - doorHitboxPadding;
      doorMaxY = doorMinY + doorHeights[curDoor.side] + doorHitboxPadding;
      // Check if "in" x-wise, then y-wise. If both, it's a collision.
      if((charMinX <= doorMaxX && charMinX >= doorMinX) || (charMaxX >= doorMinX && charMaxX <= doorMaxX) || (charMinX <= doorMinX && charMaxX >= doorMaxX)) {
        if((charMinY <= doorMaxY && charMinY >= doorMinY) || (charMaxY >= doorMinY && charMaxY <= doorMaxY) || (charMinY <= doorMinY && charMaxY >= doorMaxY)) {
          doorIndex = i;
          break;
        }
      }
    }
    return doorIndex;
  }
  
  /* Returns the X coordinate of the given door. */
  public int getDoorX(DoorInfo door) {
    int offset = door.offset;
    int doorX = 0;
    switch(door.side) {
      case 0: // North.
        doorX = (unitsX + offset) * sizeUnitScale + doorOffset - (doorWidths[door.side] / 2); 
        break;
      case 1: // South.
        doorX = (unitsX + offset) * sizeUnitScale + doorOffset - (doorWidths[door.side] / 2);
        break;
      case 2: // East.
        doorX = (unitsX + roomRef.unitWidth) * sizeUnitScale - doorWidths[door.side];
        break;
      case 3: // West.
        doorX = unitsX * sizeUnitScale;
        break;
    }
    return doorX;
  }
  
  /* Returns the Y coordinate of the given door. */
  public int getDoorY(DoorInfo door) {
    int offset = door.offset;
    int doorY = 0;
    switch(door.side) {
      case 0: // North.
        doorY = unitsY * sizeUnitScale - wallHeight + (wallHeight - doorHeights[door.side]); 
        break;
      case 1: // South.
        doorY = (unitsY + roomRef.unitHeight) * sizeUnitScale - doorHeights[door.side];
        break;
      case 2: // East.
        doorY = (unitsY + offset) * sizeUnitScale + doorOffset - (doorHeights[door.side] / 2);
        break;
      case 3: // West.
        doorY = (unitsY + offset) * sizeUnitScale + doorOffset - (doorHeights[door.side] / 2);
        break;
    }
    return doorY;
  }
}

/* Randomly selects a RoomInfo to create a new room from.
   If _needsDoorDirection is -1, any room will do. Otherwise,
   requires the room have a N (0), S (1), E (2), or W (2) door. */
public RoomInfo getNewRoom(int _needsDoorDirection) {
  boolean acceptable = false;
  RoomInfo room = roomLoader.getRoomInfo(PApplet.parseInt(random(roomLoader.size())));;
  while(!acceptable) {
    if(_needsDoorDirection >= 0) {
      for(DoorInfo door : room.doors) {
        if(door.side == _needsDoorDirection) {
          acceptable = true;
        }
      }
    } else {
      // No door requirements, so any room will do.
      acceptable = true;
    }
    if(!acceptable) {
      room = roomLoader.getRoomInfo(PApplet.parseInt(random(roomLoader.size()))); 
    }
  }
  return room;
}

/* Checks if any character is in a given room. */
public boolean isACharacterInRoom(RoomData room) {
  for(int i = 0; i < characters.size(); i++) {
    int charUnitX = PApplet.parseInt(characters.get(i).x / sizeUnitScale);
    int charUnitY = PApplet.parseInt(characters.get(i).y / sizeUnitScale);
    
    if(!characters.get(i).isDead && characters.get(i).curRoom == room) {
      return true;
      /*
      if(charUnitX >= room.unitsX && charUnitX <= room.unitsX + room.roomRef.unitWidth) {
        if(charUnitY >= room.unitsY && charUnitY <= room.unitsY + room.roomRef.unitHeight) {
          return true;
        }
      }
      */
    }
  }
  return false;
}

/* Attempts to find a room next to curRoom through the door at doorIndex.
   Returns the index of that room, or -1 if none exists yet. */
public int getNeighborRoom(RoomData curRoom, DoorInfo curDoor) {
  int neighborRoomIndex = -1;
  for(int i = 0; i < discoveredRooms.size(); i++) {
    RoomData checkRoom = discoveredRooms.get(i);
    switch(curDoor.side) {
      case 0: // North.
        if(curRoom.unitsY == checkRoom.unitsY + checkRoom.roomRef.unitHeight) {
          if(curRoom.unitsX + curDoor.offset >= checkRoom.unitsX && curRoom.unitsX + curDoor.offset <= checkRoom.unitsX + checkRoom.roomRef.unitWidth) {
            return i;
          }
        }
        break;
      case 1: // South.
        if(curRoom.unitsY + curRoom.roomRef.unitHeight == checkRoom.unitsY) {
          if(curRoom.unitsX + curDoor.offset >= checkRoom.unitsX && curRoom.unitsX + curDoor.offset <= checkRoom.unitsX + checkRoom.roomRef.unitWidth) {
            return i;
          }
        }
        break;
      case 2: // East.
        if(curRoom.unitsX + curRoom.roomRef.unitWidth == checkRoom.unitsX) {
          if(curRoom.unitsY + curDoor.offset >= checkRoom.unitsY && curRoom.unitsY + curDoor.offset <= checkRoom.unitsY + checkRoom.roomRef.unitHeight) {
            return i;
          }
        }
        break;
      case 3: // West.
        if(curRoom.unitsX == checkRoom.unitsX + checkRoom.roomRef.unitWidth) {
          if(curRoom.unitsY + curDoor.offset >= checkRoom.unitsY && curRoom.unitsY + curDoor.offset <= checkRoom.unitsY + checkRoom.roomRef.unitHeight) {
            return i;
          }
        }
        break;
    }
  }
  return neighborRoomIndex;
}

/* Attempts to find an existing room next to the given room through the given door,
   or generate an appropriate new one. Returns the index of the new room in the
   discoveredRooms array, or -1 if a room couldn't be found/generated. */
public int generateNewRoom(RoomData curRoom, DoorInfo curDoor) {
  int newRoomIndex = -1;
  int neededDoorSide = -1;
  int newX = 0;
  int newY = 0;
  
  switch(curDoor.side) {
    case 0: // North.
      neededDoorSide = 1;
      break;
    case 1: // South.
      neededDoorSide = 0;
      break;
    case 2: // East.
      neededDoorSide = 3;
      break;
    case 3:
      neededDoorSide = 2;
      break;
  }
  RoomInfo newRoomInfo = getNewRoom(neededDoorSide);
  // Set the new room's X and Y based on which door side is connecting to the current room.
  // TODO: WARNING: Assumption made here that all rooms are 2 units wide by 2 units high.
  switch(curDoor.side) {
    case 0: // New room is North.
      newX = curRoom.unitsX;
      newY = curRoom.unitsY - newRoomInfo.unitHeight;
      break;
    case 1: // New room is South.
      newX = curRoom.unitsX;
      newY = curRoom.unitsY + curRoom.roomRef.unitHeight;
      break;
    case 2: // New room is East.
      newX = curRoom.unitsX + curRoom.roomRef.unitWidth;
      newY = curRoom.unitsY;
      break;
    case 3: // New room is West.
      newX = curRoom.unitsX - newRoomInfo.unitWidth;
      newY = curRoom.unitsY;
      break;
  }
  RoomData newRoom = new RoomData(newRoomInfo, newX, newY);
  discoveredRooms.add(newRoom);
  newRoomIndex = discoveredRooms.size() - 1;
  return newRoomIndex;
}

/* Given an "exiting" room & door and the desired "entering" room,
   tries to find the appropriate door in the new room that connects to the
   door of the old room. Returns the index of the door for the new room, or
   -1 if there isn't a matching door. */
public int getNextRoomDoor(RoomData curRoom, RoomData nextRoom, DoorInfo curDoor) {
  int newDoorIndex = -1;
  ArrayList<DoorInfo> doors = nextRoom.roomRef.doors;
  DoorInfo checkDoor;
  int doorX = 0;
  int doorY = 0;
  int neededSide = -1;
  int checkDoorX = 0;
  int checkDoorY = 0;
  
  switch(curDoor.side) {
    case 0: // North.
      doorX = curRoom.unitsX + 2 * curDoor.offset;
      doorY = curRoom.unitsY;
      neededSide = 1;
      break;
    case 1: // South.
      doorX = curRoom.unitsX + 2 * curDoor.offset;
      doorY = curRoom.unitsY + curRoom.roomRef.unitHeight;
      neededSide = 0;
      break;
    case 2: // East.
      doorX = curRoom.unitsX + curRoom.roomRef.unitWidth;
      doorY = curRoom.unitsY + 2 * curDoor.offset;
      neededSide = 3;
      break;
    case 3: // West.
      doorX = curRoom.unitsX;
      doorY = curRoom.unitsY + 2 * curDoor.offset;
      neededSide = 2;
      break;
  }
  // Look through the doors of the target room trying to find a match.
  for(int i = 0; i < doors.size(); i++) {
    checkDoor = doors.get(i);
    if(checkDoor.side == neededSide) {
      switch(checkDoor.side) {
        case 0: // North.
          checkDoorX = nextRoom.unitsX + 2 * curDoor.offset;
          checkDoorY = nextRoom.unitsY;
          break;
        case 1: // South.
          checkDoorX = nextRoom.unitsX + 2 * curDoor.offset;
          checkDoorY = nextRoom.unitsY + nextRoom.roomRef.unitHeight;
          break;
        case 2: // East.
          checkDoorX = nextRoom.unitsX + nextRoom.roomRef.unitWidth;
          checkDoorY = nextRoom.unitsY + 2 * curDoor.offset;
          break;
        case 3: // West.
          checkDoorX = nextRoom.unitsX;
          checkDoorY = nextRoom.unitsY + 2 * curDoor.offset;
          break;
      }
      if(doorX == checkDoorX && doorY == checkDoorY) {
        newDoorIndex = i;
        break;
      }
    }
  }
  return newDoorIndex;
}
// How many pixels count as one "unit" on the room grid.
int sizeUnitScale = 150;

/* Loads data from a text file describing room properties. */
class RoomLoader {
  ArrayList<RoomInfo> rooms;
  
  RoomLoader(String fileName) {
    rooms = new ArrayList<RoomInfo>();
    
    String floor;
    String wall;
    String name;
    int unitWidth;
    int unitHeight;
    ArrayList<DoorInfo> doors;
    Table roomData = loadTable(fileName, "header");
    for(TableRow row : roomData.rows()) {
      floor = row.getString(0);
      wall = row.getString(1);
      name = row.getString(2);
      unitWidth = row.getInt(3);
      unitHeight = row.getInt(4);
      // Grab door infornation until the row runs out.
      doors = new ArrayList<DoorInfo>();
      try {
        int i = 5;
        String doorSide;
        int doorOffset;
        while(true) {
          doorSide = row.getString(i);
          i++;
          doorOffset = row.getInt(i);
          i++;
          doors.add(new DoorInfo(doorSide, doorOffset));
        }
      } catch(Exception e) {
        // Do nothing, just don't want to crash.
      }
      rooms.add(new RoomInfo(floor, wall, name, unitWidth, unitHeight, doors));
    }
  }
  
  /* Creates a short text description of a room at the given index. */
  public String getRoomSummary(int index) {
    String summary = rooms.get(index).name + ": ";
    summary = summary + (rooms.get(index).unitWidth * sizeUnitScale) + " x " + (rooms.get(index).unitHeight * sizeUnitScale);
    summary = summary + ", Doors:";
    for(int i = 0; i < rooms.get(index).doors.size(); i++) {
      summary = summary + " " + rooms.get(index).doors.get(i).getDoorSideName() + "/" + rooms.get(index).doors.get(i).offset;
    }
    return summary;
  }
  
  public RoomInfo getRoomInfo(int index) {
    return rooms.get(index);
  }
  
  public int size() {
    return rooms.size();
  }
}

/* Container class to hold information about a room pulled from the room definitions. */
class RoomInfo {
  String floorPicture;
  String wallPicture;
  String name;
  int unitWidth;
  int unitHeight;
  ArrayList<DoorInfo> doors;
  PImage floorImage;
  PImage wallImage;
  
  RoomInfo(String _floor, String _wall, String _name, int _w, int _h, ArrayList<DoorInfo> _doors) {
    floorPicture = _floor;
    wallPicture = _wall;
    name = _name;
    unitWidth = _w;
    unitHeight = _h;
    doors = _doors;
    
    floorImage = loadImage(_floor);
    wallImage = loadImage(_wall);
  }
}

/* Container class to hold info about a door's position in a room. */
class DoorInfo {
  int side; // "N" = 0, "S" = 1, "E" = 2, "W" = 3.
  int offset; // 0 = 1st possible left/top door position, 1 = 2nd, etc.
  DoorInfo(String _side, int _offset) {
    String sideStr = _side.toUpperCase();
    if(sideStr.equals("N")) {
      side = 0;
    } else if(sideStr.equals("S")) {
      side = 1;
    } else if(sideStr.equals("E")) {
      side = 2;
    } else {
      side = 3;
    }
    offset = _offset;
  }
  
  public String getDoorSideName() {
    if(side == 0) {
      return "N";
    } else if(side == 1) {
      return "S";
    } else if(side == 2) {
      return "E";
    } else {
      return "W";
    }
  }
}
int mainTextBoxLeft = 10;
int mainTextBoxTop = 640;
int mainTextBoxWidth = 1100;
int mainTextBoxHeight = 40;
int textBoxPadding = 4;
int mainTextBoxStroke = color(0, 255, 255);
int mainTextBoxFill = color(128, 255, 255, 128);

int mainFontSize = 24;
int normalColor = color(0, 0, 32);
int scaryColor = color(64, 0, 16);

PFont mainTextFont = createFont("Georgia", mainFontSize, true);

/* A UI element to display text in a box. */
class TextBox {
  int x;
  int y;
  int w;
  int h;
  int fontSize;
  int strokeCol;  // The color of the border of the text box.
  int fillCol;  // The color of the "background" of the text box.
  
  TextBox(int _x, int _y, int _w, int _h, int _fontSize, int _strokeCol, int _fillCol) {
    x = _x;
    y = _y;
    w = _w;
    h = _h;
    fontSize = _fontSize;
    strokeCol = _strokeCol;
    fillCol = _fillCol;
  }
  
  /* Displays the passed text in a text box. */
  public void displayText(String dispText, int col) {
    // Draw the text box.
    tint(255);
    stroke(strokeCol);
    fill(fillCol);
    rect(x, y, w, h);
    // Render the text on top of the text box.
    textFont(mainTextFont, fontSize);
    fill(col);
    text(dispText, x + textBoxPadding, y + fontSize + textBoxPadding);
  }
}

/* Subclass of TextBox intended to be the "main" info text box of the GUI. */
class MainTextBox extends TextBox {
  PImage screenSave;  // Used to preserve what the window looked like before the box popped up.
  boolean displayingText;
  String currentText;
  int currentCol;
  
  MainTextBox(int _x, int _y, int _w, int _h, int _fontSize, int _strokeCol, int _fillCol) {
    super(_x, _y, _w, _h, _fontSize, _strokeCol, _fillCol);
    screenSave = createImage(width, height, RGB);
    displayingText = false;
    currentText = "";
    currentCol = color(0, 0, 0);
  }
  
  /* Pops up the passed text in a text box. */
  public void displayText(String dispText, int col) {
    // Only save the background if this is the first frame this text box is visible.
    if(displayingText == false) {
      savePixels(screenSave);
    }
    currentText = dispText;
    currentCol = col;
    displayingText = true;
    redraw();  // Easier to just call this rather than repeating the code here.
  }
  
  /* Re-renders the text box without changing anything. */
  public void redraw() {
    // Draw the text box.
    tint(255);
    stroke(strokeCol);
    fill(fillCol);
    rect(x, y, w, h, 7);
    // Render the text on top of the text box.
    textFont(mainTextFont, fontSize);
    fill(currentCol);
    text(currentText, x + textBoxPadding, y + fontSize + textBoxPadding);
  }
  
  /* Closes the text box and restores what was underneath. */
  public void closeText() {
    restorePixels(screenSave);
    displayingText = false;
  }
  
}

/* Saves off the current screen so it can be restored later. */
public void savePixels(PImage buffer) {
  buffer.loadPixels();
  loadPixels();
  int loc;
  for(int x = 0; x < width; x++) {
    for(int y = 0; y < height; y++) {
      loc = x + y * width;
      buffer.pixels[loc] = pixels[loc];
    }
  }
  updatePixels();
  buffer.updatePixels();
}

/* Restores the screen to what was saved in a PImage. */
public void restorePixels(PImage buffer) {
  buffer.loadPixels();
  loadPixels();
  int loc;
  for(int x = 0; x < width; x++) {
    for(int y = 0; y < height; y++) {
      loc = x + y * width;
      pixels[loc] = buffer.pixels[loc];
    }
  }
  updatePixels();
  buffer.updatePixels();
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "--full-screen", "--bgcolor=#666666", "--hide-stop", "Main" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
