/* Loads data from a text file describing possible party members. */
class PartyLoader {
  ArrayList<PartyMemberInfo> partyInfo;
  
  PartyLoader(String fileName) {
    partyInfo = new ArrayList<PartyMemberInfo>();
    
    Table partyData = loadTable(fileName, "header");
    for(TableRow row : partyData.rows()) {
      String name = row.getString(0);
      String genderStr = row.getString(1);
      String healthyPicture = row.getString(2);
      String hurtPicture = row.getString(3);
      String deadPicture = row.getString(4);
      String spritePicture = row.getString(5);
      int maxHealth = row.getInt(6);
      int maxMovement = row.getInt(7);
      partyInfo.add(new PartyMemberInfo(name, genderStr, healthyPicture, hurtPicture, deadPicture, spritePicture, maxHealth, maxMovement));
    }
  }
  
  /* Creates a short text description of the party member. */
  String getPartyMemberSummary(int index) {
    return partyInfo.get(index).getPartyMemberSummary();
  }
  
  PartyMemberInfo getPartyMember(int index) {
    return partyInfo.get(index);
  }
  
  int size() {
    return partyInfo.size();
  }
}

/* Container class for information loaded about a party member. */
class PartyMemberInfo {
  String name;
  int gender; // 0 = Male ("M"), 1 = Female ("F").
  String healthyPicture;
  String hurtPicture;
  String deadPicture;
  String spritePicture;
  int maxHealth;
  int maxMovement;
  
  PartyMemberInfo(String _name, String _gender, String _healthyPicture, String _hurtPicture, String _deadPicture, String _spritePicture, int _maxHealth, int _maxMovement) {
    name = _name;
    if(_gender.toLowerCase().equals("m")) {
      gender = 0;
    } else {
      gender = 1;
    }
    healthyPicture = _healthyPicture;
    hurtPicture = _hurtPicture;
    deadPicture = _deadPicture;
    spritePicture = _spritePicture;
    maxHealth = _maxHealth;
    maxMovement = _maxMovement;
  }
  
  /* Returns a human-readable gender for the party member. */
  String getGenderName() {
    if(gender == 0) {
      return "Male";
    } else {
      return "Female";
    }
  }
  
  /* Creates a short text description of the party member. */
  String getPartyMemberSummary() {
    String summary;
    summary = name;
    summary = summary + " (" + getGenderName() + ")";
    summary = summary + ", Max Health " + maxHealth;
    summary = summary + ", Speed " + maxMovement;
    return summary;
  }
}
