// How many pixels count as one "unit" on the room grid.
int sizeUnitScale = 150;

/* Loads data from a text file describing room properties. */
class RoomLoader {
  ArrayList<RoomInfo> rooms;
  
  RoomLoader(String fileName) {
    rooms = new ArrayList<RoomInfo>();
    
    String floor;
    String wall;
    String name;
    int unitWidth;
    int unitHeight;
    ArrayList<DoorInfo> doors;
    Table roomData = loadTable(fileName, "header");
    for(TableRow row : roomData.rows()) {
      floor = row.getString(0);
      wall = row.getString(1);
      name = row.getString(2);
      unitWidth = row.getInt(3);
      unitHeight = row.getInt(4);
      // Grab door infornation until the row runs out.
      doors = new ArrayList<DoorInfo>();
      try {
        int i = 5;
        String doorSide;
        int doorOffset;
        while(true) {
          doorSide = row.getString(i);
          i++;
          doorOffset = row.getInt(i);
          i++;
          doors.add(new DoorInfo(doorSide, doorOffset));
        }
      } catch(Exception e) {
        // Do nothing, just don't want to crash.
      }
      rooms.add(new RoomInfo(floor, wall, name, unitWidth, unitHeight, doors));
    }
  }
  
  /* Creates a short text description of a room at the given index. */
  String getRoomSummary(int index) {
    String summary = rooms.get(index).name + ": ";
    summary = summary + (rooms.get(index).unitWidth * sizeUnitScale) + " x " + (rooms.get(index).unitHeight * sizeUnitScale);
    summary = summary + ", Doors:";
    for(int i = 0; i < rooms.get(index).doors.size(); i++) {
      summary = summary + " " + rooms.get(index).doors.get(i).getDoorSideName() + "/" + rooms.get(index).doors.get(i).offset;
    }
    return summary;
  }
  
  RoomInfo getRoomInfo(int index) {
    return rooms.get(index);
  }
  
  int size() {
    return rooms.size();
  }
}

/* Container class to hold information about a room pulled from the room definitions. */
class RoomInfo {
  String floorPicture;
  String wallPicture;
  String name;
  int unitWidth;
  int unitHeight;
  ArrayList<DoorInfo> doors;
  PImage floorImage;
  PImage wallImage;
  
  RoomInfo(String _floor, String _wall, String _name, int _w, int _h, ArrayList<DoorInfo> _doors) {
    floorPicture = _floor;
    wallPicture = _wall;
    name = _name;
    unitWidth = _w;
    unitHeight = _h;
    doors = _doors;
    
    floorImage = loadImage(_floor);
    wallImage = loadImage(_wall);
  }
}

/* Container class to hold info about a door's position in a room. */
class DoorInfo {
  int side; // "N" = 0, "S" = 1, "E" = 2, "W" = 3.
  int offset; // 0 = 1st possible left/top door position, 1 = 2nd, etc.
  DoorInfo(String _side, int _offset) {
    String sideStr = _side.toUpperCase();
    if(sideStr.equals("N")) {
      side = 0;
    } else if(sideStr.equals("S")) {
      side = 1;
    } else if(sideStr.equals("E")) {
      side = 2;
    } else {
      side = 3;
    }
    offset = _offset;
  }
  
  String getDoorSideName() {
    if(side == 0) {
      return "N";
    } else if(side == 1) {
      return "S";
    } else if(side == 2) {
      return "E";
    } else {
      return "W";
    }
  }
}
